﻿using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

namespace StereoKit
{
    /// <summary> Bounds is an axis aligned bounding box type that can be used for
    /// storing the sizes of objects, calculating containment, intersections, and
    /// more!
    /// 
    /// While the constructor uses a center+dimensions for creating a bounds, don't 
    /// forget the static From* methods that allow you to define a Bounds from different
    /// types of data!</summary>
    [StructLayout(LayoutKind.Sequential)]
    public struct Bounds
    {
        /// <summary>The exact center of the Bounds!</summary>
        public Vec3 center;
        /// <summary>The total size of the box, from one end to the other. This is the 
        /// width, height, and depth of the Bounds.</summary>
        public Vec3 dimensions;

        /// <summary>Creates a bounding box object!</summary>
        /// <param name="center">The exact center of the box.</param>
        /// <param name="totalDimensions">The total size of the box, from one end to the 
        /// other. This is the width, height, and depth of the Bounds.</param>
        public Bounds(Vec3 center, Vec3 totalDimensions)
        {
            this.center = center;
            dimensions = totalDimensions;
        }

        /// <summary>Creates a bounding box object centered around zero!</summary>
        /// <param name="totalDimensions">The total size of the box, from one end to the 
        /// other. This is the width, height, and depth of the Bounds.</param>
        public Bounds(Vec3 totalDimensions)
        {
            this.center = Vec3.Zero;
            dimensions = totalDimensions;
        }

        /// <summary>Create a bounding box from a corner, plus box dimensions.</summary>
        /// <param name="bottomLeftBack">The -X,-Y,-Z corner of the box.</param>
        /// <param name="dimensions">The total dimensions of the box.</param>
        /// <returns>A Bounds object that extends from bottomLeftBack to bottomLeftBack+dimensions.</returns>
        public static Bounds FromCorner(Vec3 bottomLeftBack, Vec3 dimensions)
            => new Bounds(bottomLeftBack + dimensions/2, dimensions);

        /// <summary>Create a bounding box between two corner points.</summary>
        /// <param name="bottomLeftBack">The -X,-Y,-Z corner of the box.</param>
        /// <param name="topRightFront">The +X,+Y,+Z corner of the box.</param>
        /// <returns>A Bounds object that extends from bottomLeftBack to topRightFront.</returns>
        public static Bounds FromCorners(Vec3 bottomLeftBack, Vec3 topRightFront)
            => new Bounds(bottomLeftBack/2 + topRightFront/2, topRightFront-bottomLeftBack);
        

        /// <summary>Calculate the intersection between a Ray, and these bounds. Returns false
        /// if no intersection occurred, and 'at' will contain the nearest intersection point to 
        /// the start of the ray if an intersection is found!</summary>
        /// <param name="ray">Any Ray in the same coordinate space as the Bounds.</param>
        /// <param name="at">If the return is true, this point will be the closest intersection 
        /// point to the origin of the Ray.</param>
        /// <returns>True if an intersection occurred, false if not.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public bool Intersect(Ray ray, out Vec3 at)
            => NativeAPI.bounds_ray_intersect(this, ray, out at);

        /// <summary>Does the Bounds contain the given point? This includes points that are on
        /// the surface of the Bounds.</summary>
        /// <param name="pt">A point in the same coordinate space as the Bounds.</param>
        /// <returns>True if the point is on, or in the Bounds.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public bool Contains(Vec3 pt)
            => NativeAPI.bounds_point_contains(this, pt);

        /// <summary>Does the Bounds contain or intersects with the given line?</summary>
        /// <param name="linePt1">Start of the line</param>
        /// <param name="linePt2">End of the line</param>
        /// <returns>True if the line is in, or intersects with the bounds.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public bool Contains(Vec3 linePt1, Vec3 linePt2)
            => NativeAPI.bounds_line_contains(this, linePt1, linePt2);

        public static Bounds operator *(Bounds a, float b) { return new Bounds(a.center*b, a.dimensions*b); }

        public override string ToString()
        {
            return string.Format("[center:{0} dimensions:{1}", center, dimensions);
        }
    }
}
